#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, List, Optional, TYPE_CHECKING

import dateutil
import trafaret as t
from trafaret import Bool, Float

from datarobot._compat import String
from datarobot.enums import BUCKET_SIZE
from datarobot.models.api_object import APIObject
from datarobot.models.deployment.mixins import MonitoringDataQueryBuilderMixin
from datarobot.utils import from_api

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    class Period(TypedDict, total=False):
        start: datetime
        end: datetime

    class Bucket(TypedDict):
        period: Period
        value: int
        sample_size: int


class FairnessScoresOverTime(APIObject, MonitoringDataQueryBuilderMixin):
    """Deployment fairness over time information.

    Attributes
    ----------
    buckets : List
        fairness over time bucket for each model and bucket queried
    summary : dict
        summary for the fairness score
    protected_feature : str
        name of protected feature
    fairnessThreshold : float
        threshold used to compute fairness results
    modelId : str
        model id for which fairness is computed
    modelPackageId : str
        model package (version) id for which fairness is computed
    favorableTargetOutcome : bool
        preferable class of the target
    """

    _path = "deployments/{}/fairnessScoresOverTime/"
    _period = t.Dict(
        {
            t.Key("start"): String >> dateutil.parser.parse,
            t.Key("end"): String >> dateutil.parser.parse,
        }
    )
    _bucket = t.Dict(
        {
            t.Key("period"): t.Or(_period, t.Null),
            t.Key("metric_name"): t.Or(t.String(), t.Null),
            t.Key("scores"): t.Or(
                t.List(
                    t.Dict(
                        {
                            t.Key("label"): t.String(),
                            t.Key("absolute_value"): t.Int(),
                            t.Key("classes_count"): t.Int(),
                            t.Key("healthy_classes_count"): t.Int(),
                            t.Key("is_statistically_significant"): t.Bool(),
                            t.Key("priviledged_class"): t.String(),
                            t.Key("sample_size"): t.Int(),
                            t.Key("value"): t.Int(),
                        }
                    )
                ),
                t.Null,
            ),
        }
    ).allow_extra("*")
    _converter = t.Dict(
        {
            t.Key("protected_feature"): t.Or(String(), t.Null),
            t.Key("fairness_threshold"): t.Or(Float(), t.Null),
            t.Key("model_id"): t.Or(String(), t.Null),
            t.Key("model_package_id"): t.Or(String(), t.Null),
            t.Key("favorable_target_outcome"): t.Or(Bool(), t.Null),
            t.Key("summary"): _bucket,
            t.Key("buckets"): t.List(t.Dict({"period": _period}).allow_extra("*")),
        }
    )

    def __init__(
        self,
        summary: Optional[Dict[str, Any]] = None,
        buckets: Optional[List[Bucket]] = None,
        protected_feature: Optional[str] = None,
        fairness_threshold: Optional[float] = None,
        model_id: Optional[str] = None,
        model_package_id: Optional[str] = None,
        favorable_target_outcome: Optional[bool] = None,
    ):
        self.summary = summary
        self.buckets = buckets if buckets is not None else []
        self.protected_feature = protected_feature
        self.fairness_threshold = fairness_threshold
        self.model_id = model_id
        self.model_package_id = model_package_id
        self.favorable_target_outcome = favorable_target_outcome

    @classmethod
    def get(
        cls,
        deployment_id: str,
        model_id: Optional[str] = None,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        bucket_size: Optional[BUCKET_SIZE] = None,
        fairness_metric: Optional[str] = None,
        protected_feature: Optional[str] = None,
    ) -> FairnessScoresOverTime:
        """Retrieve information for deployment's fairness score response over a certain time period.

        .. versionadded:: FUTURE

        Parameters
        ----------
        deployment_id : str
            the id of the deployment
        model_id : str
            id of models to retrieve fairness score stats
        start_time : datetime
            start of the time period
        end_time : datetime
            end of the time period
        protected_feature : str
            name of the protected feature
        fairness_metric : str
            A consolidation of the fairness metrics by the use case.
        bucket_size : BUCKET_SIZE
            time duration of each bucket

        Returns
        -------
        fairness_scores_over_time : FairnessScoresOverTime
            the queried fairness score over time information
        """

        path = cls._path.format(deployment_id)
        params = cls._build_query_params(
            start_time=start_time,
            end_time=end_time,
            model_id=model_id,
            bucket_size=bucket_size,
            fairness_metric=fairness_metric,
            protected_feature=protected_feature,
        )
        data = cls._client.get(path, params=params).json()
        case_converted = from_api(data, keep_null_keys=True)
        return cls.from_data(case_converted)



#
# Copyright 2021-2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from enum import Enum
from typing import cast, List, Optional, Set

import trafaret as t

from datarobot._compat import String
from datarobot.enums import DEFAULT_MAX_WAIT
from datarobot.errors import ClientError
from datarobot.models.api_object import APIObject
from datarobot.models.registry.job import Job, JobFileItem, JobFileItemType
from datarobot.models.runtime_parameters import RuntimeParameter, RuntimeParameterValue
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution


class JobRunStatus(Enum):
    """Enum of the job run statuses"""

    SUCCEEDED = "succeeded"
    FAILED = "failed"
    RUNNING = "running"
    INTERRUPTED = "interrupted"
    CANCELING = "canceling"
    CANCELED = "canceled"


class JobRun(APIObject):
    """A DataRobot job run.

    .. versionadded:: v3.4

    Attributes
    ----------
    id: str
        The ID of the job run.
    custom_job_id: str
        The ID of the parent job.
    description: str
        A description of the job run.
    created_at: str
        ISO-8601 formatted timestamp of when the version was created
    items: List[JobFileItem]
        A list of file items attached to the job.
    status: JobRunStatus
        The status of the job run.
    duration: float
        The duration of the job run.
    """

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("custom_job_id"): String(),
            t.Key("description", optional=True): t.Or(
                String(max_length=10000, allow_blank=True), t.Null()
            ),
            t.Key("created") >> "created_at": String(),
            t.Key("items"): t.List(JobFileItem.schema),
            t.Key("status"): t.Enum(*[e.value for e in JobRunStatus]),
            t.Key("duration"): t.Float(),
            t.Key("runtime_parameters", optional=True): t.List(RuntimeParameter.schema),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(
        self,
        id: str,
        custom_job_id: str,
        created_at: str,
        items: List[JobFileItemType],
        status: str,
        duration: float,
        description: Optional[str] = None,
        runtime_parameters: Optional[List[RuntimeParameter]] = None,
    ) -> None:
        self.id = id
        self.custom_job_id = custom_job_id
        self.description = description
        self.created_at = created_at

        # NOTE: JobFileItem's __init__ instead of from_server_data is used, because at this point
        #   the data is already converted from API representation to "object" representation.
        #   In case of JobFileItem it converted {"created": ..., ...} to {"created_at": ..., ...}.
        #   For type hinting, TypedDict JobFileItemType is used, which reflects expected property names and types.
        self.items = [JobFileItem(**data) for data in items]

        self.status = JobRunStatus(status)
        self.duration = duration

        self.runtime_parameters = (
            [RuntimeParameter(**param) for param in runtime_parameters]  # type: ignore[arg-type]
            if runtime_parameters
            else None
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.id!r})"

    def _update_values(self, new_response: JobRun) -> None:
        fields: Set[str] = self._fields()  # type: ignore[no-untyped-call]
        for attr in fields:
            new_value = getattr(new_response, attr)
            setattr(self, attr, new_value)

    @classmethod
    def _job_run_base_path(cls, job_id: str) -> str:
        return f"customJobs/{job_id}/runs/"

    @classmethod
    def _job_run_path(cls, job_id: str, job_run_id: str) -> str:
        return f"customJobs/{job_id}/runs/{job_run_id}/"

    @classmethod
    def _job_run_logs_path(cls, job_id: str, job_run_id: str) -> str:
        return f"customJobs/{job_id}/runs/{job_run_id}/logs/"

    @classmethod
    def create(
        cls,
        job_id: str,
        max_wait: Optional[int] = DEFAULT_MAX_WAIT,
        runtime_parameter_values: Optional[List[RuntimeParameterValue]] = None,
    ) -> JobRun:
        """Create a job run.

        .. versionadded:: v3.4

        Parameters
        ----------
        job_id: str
            The ID of the job.
        max_wait: int, optional
            max time to wait for a terminal status ("succeeded", "failed", "interrupted", "canceled").
            If set to None - method will return without waiting.
        runtime_parameter_values: Optional[List[RuntimeParameterValue]]
            Additional parameters to be injected into a model at runtime. The fieldName
            must match a fieldName that is listed in the runtimeParameterDefinitions section
            of the model-metadata.yaml file.

        Returns
        -------
        Job
            created job

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        ValueError
            if execution environment or entry point is not specified for the job
        """

        job = Job.get(job_id)
        if not job.environment_id:
            raise ValueError("Environment ID must be set for the job in order to be run")
        if not job.entry_point:
            raise ValueError("Entry point must be set for the job in order to be run")

        path = cls._job_run_base_path(job_id)

        payload = {}

        if runtime_parameter_values:
            payload["runtimeParameterValues"] = [
                param.to_dict() for param in runtime_parameter_values
            ]

        response = cls._client.post(path, data=payload)

        data = response.json()

        if max_wait is None:
            return cls.from_server_data(data)

        job_run_loc = wait_for_async_resolution(cls._client, response.headers["Location"], max_wait)
        return cls.from_location(job_run_loc)

    @classmethod
    def list(cls, job_id: str) -> List[JobRun]:
        """List job runs.

        .. versionadded:: v3.4

        Parameters
        ----------
        job_id: str
            The ID of the job.

        Returns
        -------
        List[Job]
            A list of job runs.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        data = unpaginate(cls._job_run_base_path(job_id), None, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def get(cls, job_id: str, job_run_id: str) -> JobRun:
        """Get job run by id.

        .. versionadded:: v3.4

        Parameters
        ----------
        job_id: str
            The ID of the job.
        job_run_id: str
            The ID of the job run.

        Returns
        -------
        Job
            The retrieved job run.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        path = cls._job_run_path(job_id, job_run_id)
        return cls.from_location(path)

    def update(self, description: Optional[str] = None) -> None:
        """Update job run properties.

        .. versionadded:: v3.4

        Parameters
        ----------
        description: str
            new job run description

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        payload = {}
        if description:
            payload.update({"description": description})

        path = self._job_run_path(self.custom_job_id, self.id)

        response = self._client.patch(path, data=payload)

        data = response.json()
        new_version = JobRun.from_server_data(data)
        self._update_values(new_version)

    def cancel(self) -> None:
        """Cancel job run.

        .. versionadded:: v3.4

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        url = self._job_run_path(self.custom_job_id, self.id)
        self._client.delete(url)

    def refresh(self) -> None:
        """Update job run with the latest data from server.

        .. versionadded:: v3.4

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """

        new_object = self.get(self.custom_job_id, self.id)
        self._update_values(new_object)

    def get_logs(self) -> Optional[str]:
        """Get log of the job run.

        .. versionadded:: v3.4

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        path = self._job_run_logs_path(self.custom_job_id, self.id)
        try:
            response = self._client.get(path)
            return cast(str, response.text)
        except ClientError as exc:
            if exc.status_code == 404 and exc.json == {"message": "No log found"}:
                return None
            raise

    def delete_logs(self) -> None:
        """Get log of the job run.

        .. versionadded:: v3.4

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        path = self._job_run_logs_path(self.custom_job_id, self.id)
        self._client.delete(path)



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, Iterator, List, TYPE_CHECKING

import trafaret as t

from datarobot._compat import Int, String
from datarobot.models.api_object import APIObject

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    class TaskInfo(TypedDict):
        name: str
        descriptions: List[str]

    class MissingReportPerFeatureDict(TypedDict):
        feature: str
        type: str
        missing_count: int
        missing_percentage: float
        tasks: Dict[str, TaskInfo]


class MissingValuesReport(APIObject):
    """Missing values report for model, contains list of reports per feature sorted by missing
    count in descending order.

    Notes
    -----

    ``Report per feature`` contains:

    * ``feature`` : feature name.
    * ``type`` : feature type -- 'Numeric' or 'Categorical'.
    * ``missing_count`` :  missing values count in training data.
    * ``missing_percentage`` : missing values percentage in training data.
    * ``tasks`` : list of information per each task, which was applied to feature.

    ``task information`` contains:

    * ``id`` : a number of task in the blueprint diagram.
    * ``name`` : task name.
    * ``descriptions`` : human readable aggregated information about how the task handles
      missing values.  The following descriptions may be present: what value is imputed for
      missing values, whether the feature being missing is treated as a feature by the task,
      whether missing values are treated as infrequent values,
      whether infrequent values are treated as missing values,
      and whether missing values are ignored.

    """

    _converter = t.Dict(
        {
            t.Key("missing_values_report"): t.List(
                t.Dict(
                    {
                        "feature": String,
                        "type": String,
                        "missing_count": Int,
                        "missing_percentage": t.Float(),
                        "tasks": t.Mapping(
                            String,
                            t.Dict({"name": String, "descriptions": t.List(String)}).ignore_extra(
                                "*"
                            ),
                        ),
                    }
                ).ignore_extra("*")
            )
        }
    ).ignore_extra("*")

    def __init__(self, missing_values_report: List[MissingReportPerFeatureDict]) -> None:
        self._reports_per_feature = [
            MissingReportPerFeature(data) for data in missing_values_report
        ]

    def __iter__(self) -> Iterator[MissingReportPerFeature]:
        return iter(self._reports_per_feature)

    @classmethod
    def get(cls, project_id: str, model_id: str) -> MissingValuesReport:
        """Retrieve a missing report.

        Parameters
        ----------
        project_id : str
            The project's id.
        model_id : str
            The model's id.

        Returns
        -------
        MissingValuesReport
            The queried missing report.
        """
        url = f"projects/{project_id}/models/{model_id}/missingReport/"
        return cls.from_location(url)


class TaskMissingReportInfo:
    """Convert dictionary into object, which contains task name, id (a number of task in the
    blueprint diagram) and descriptions (task specific missing report info).
    """

    def __init__(self, task_id: str, info: TaskInfo) -> None:
        self.id = task_id
        self.name = info["name"]
        self.descriptions = info["descriptions"]

    def __repr__(self) -> str:
        return "TaskMissingReportInfo(id={}, name={}, descriptions={})".format(
            self.id, self.name, self.descriptions
        )

    def __eq__(self, other: Any) -> bool:
        return all(
            [self.id == other.id, self.name == other.name, self.descriptions == other.descriptions]
        )


class MissingReportPerFeature:
    """Convert dictionary into report per feature, which contains feature, type, missing count,
    percentage and tasks.
    """

    def __init__(self, report_per_feature_dict: MissingReportPerFeatureDict) -> None:
        self.feature = report_per_feature_dict["feature"]
        self.type = report_per_feature_dict["type"]
        self.missing_count = report_per_feature_dict["missing_count"]
        self.missing_percentage = report_per_feature_dict["missing_percentage"] * 100
        self.tasks = [
            TaskMissingReportInfo(task_id, task_info)
            for task_id, task_info in report_per_feature_dict["tasks"].items()
        ]

    def __repr__(self) -> str:
        return (
            "MissingReportPerFeature(feature={},"
            "type={}, miss_count={}, miss_percentage={}, tasks={}"
        ).format(self.feature, self.type, self.missing_count, self.missing_percentage, self.tasks)

    def __eq__(self, other: Any) -> bool:
        return all(
            [
                self.feature == other.feature,
                self.type == other.type,
                self.missing_count == other.missing_count,
                self.missing_percentage == other.missing_percentage,
                self.tasks == other.tasks,
            ]
        )



#
# Copyright 2021-2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, cast, Optional, Union

from mypy_extensions import TypedDict
import trafaret as t

from datarobot._compat import String
from datarobot.models.api_object import APIObject


class RuntimeParameterValueDict(TypedDict):
    field_name: str
    type: str
    value: Optional[Union[str, float, bool]]


class RuntimeParameter(APIObject):
    """Definition of a runtime parameter used for the custom model version, it includes
     the override value if provided.

    .. versionadded:: v3.4.0

    Attributes
    ----------
    field_name: str
        The runtime parameter name. This value is added as an environment variable when
        running custom models.
    type: str
        The value type accepted by the runtime parameter.
    description: str
        Describes how the runtime parameter impacts the running model.
    allow_empty: bool
        Indicates if the runtime parameter must be set before registration.
    min_value: float
        The minimum value for a numeric field.
    max_value: float
        The maximum value for a numeric field.
    default_value: str, bool, float or None
        The default value for the given field.
    override_value: str, bool, float or None
        The value set by the user that overrides the default set in the runtime parameter
        definition.
    current_value: str, bool, float or None
        After the default and the override values are applied, this is the value of the
        runtime parameter.
    credential_type: str
        Describes the type of credential, used only for credentials parameters.
    """

    _converter = t.Dict(
        {
            t.Key("field_name"): String(),
            t.Key("type"): String(),
            t.Key("description", optional=True): String(allow_blank=True),
            t.Key("allow_empty", optional=True): t.Bool(),
            t.Key("min_value", optional=True): t.Float(),
            t.Key("max_value", optional=True): t.Float(),
            t.Key("credential_type", optional=True): String(),
            t.Key("default_value", optional=True): t.Or(t.Bool(), t.Float(), String(), t.Null()),
            t.Key("override_value", optional=True): t.Or(t.Bool(), t.Float(), String(), t.Null()),
            t.Key("current_value", optional=True): t.Or(t.Bool(), t.Float(), String(), t.Null()),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(self, **kwargs: Any) -> None:
        self._set_values(**kwargs)

    def _set_values(
        self,
        field_name: str,
        type: str,
        description: Optional[str] = None,
        allow_empty: Optional[bool] = True,
        min_value: Optional[float] = None,
        max_value: Optional[float] = None,
        credential_type: Optional[str] = None,
        default_value: Optional[Union[float, bool, str]] = None,
        override_value: Optional[Union[float, bool, str]] = None,
        current_value: Optional[Union[float, bool, str]] = None,
    ) -> None:
        """Set values for runtime parameter"""
        self.field_name = field_name
        self.type = type
        self.description = description
        self.allow_empty = allow_empty
        self.min_value = min_value
        self.max_value = max_value
        self.credential_type = credential_type
        self.default_value = default_value
        self.override_value = override_value
        self.current_value = current_value


class RuntimeParameterValue(APIObject):
    """The definition of a runtime parameter value used for the custom model version, this defines
    the runtime parameter override.

    .. versionadded:: v3.4.0

    Attributes
    ----------
    field_name: str
        The runtime parameter name. This value is added as an environment variable when
        running custom models.
    type: str
        The value type accepted by the runtime parameter.
    value: str, bool or float
        After the default and the override values are applied, this is the value of the
        runtime parameter.
    """

    _converter = t.Dict(
        {
            t.Key("field_name"): String(),
            t.Key("type"): String(),
            t.Key("value"): t.Or(t.Bool(), t.Float(), String(), t.Null()),
        }
    )

    schema = _converter

    def __init__(self, **kwargs: Any) -> None:
        self._set_values(**kwargs)

    def __repr__(self) -> str:
        return "{}(field_name={!r}, type={!r} value={!r})".format(
            self.__class__.__name__,
            self.field_name,
            self.type,
            self.value,
        )

    def _set_values(self, field_name: str, type: str, value: Union[None, float, str, bool]) -> None:
        self.field_name = field_name
        self.type = type
        self.value = value

    def to_dict(self) -> RuntimeParameterValueDict:
        return cast(
            RuntimeParameterValueDict,
            self._converter.check(
                {"field_name": self.field_name, "type": self.type, "value": self.value}
            ),
        )



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Dict, List, Optional, TYPE_CHECKING

import trafaret as t

from datarobot._compat import String
from datarobot.models.api_object import APIObject

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    class FeatureListType(TypedDict):
        featurelist_id: str
        title: str
        has_fam: bool


class FeatureAssociationFeaturelists(APIObject):
    """
    Featurelists with feature association matrix availability flags for a project.

    Attributes
    ----------
    project_id : str
        Id of the project that contains the requested associations.
    featurelists : list fo dict
        The featurelists with the `featurelist_id`, `title` and the `has_fam` flag.
    """

    _path = "projects/{}/featureAssociationFeaturelists/"
    _converter = t.Dict(
        {
            t.Key("featurelists"): t.List(
                t.Dict(
                    {
                        t.Key("featurelist_id"): String(),
                        t.Key("title"): String(),
                        t.Key("has_fam"): t.Bool(),
                    }
                )
            )
        }
    )

    def __init__(
        self,
        project_id: Optional[str] = None,
        featurelists: Optional[List[FeatureListType]] = None,
    ) -> None:
        self.project_id = project_id
        self.featurelists = featurelists

    def __repr__(self) -> str:
        return "{}(project_id={}, featurelists={})".format(
            self.__class__.__name__, self.project_id, self.featurelists
        )

    @classmethod
    def get(cls, project_id: str) -> FeatureAssociationFeaturelists:
        """
        Get featurelists with feature association status for each.

        Parameters
        ----------
        project_id : str
             Id of the project of interest.

        Returns
        -------
        FeatureAssociationFeaturelists
            Featurelist with feature association status for each.
        """
        url = cls._path.format(project_id)
        response = cls._client.get(url)
        fam_featurelists = cls.from_server_data(response.json())
        fam_featurelists.project_id = project_id
        return fam_featurelists

    def to_dict(self) -> Dict[str, Optional[List[FeatureListType]]]:
        return {"featurelists": self.featurelists}



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import datetime
from typing import List, Optional

import dateutil
from mypy_extensions import TypedDict
import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.models.model_registry.common import UserMetadata


class DeploymentPredictionEnvironment(TypedDict):
    id: Optional[str]
    name: str
    plugin: Optional[str]
    platform: Optional[str]
    supported_model_formats: Optional[List[str]]
    is_managed_by_management_agent: Optional[bool]


class VersionAssociatedDeployment(APIObject):
    """
    Represents a deployment associated with a registered model version.

    Parameters
    ----------
    id : str
        The ID of the deployment.
    currently_deployed : bool
        Whether this version is currently deployed.
    registered_model_version : int
        The version of the registered model associated with this deployment.
    is_challenger : bool
        Whether the version associated with this deployment is a challenger.
    status : str
        The status of the deployment.
    label : str, optional
        The label of the deployment.
    first_deployed_at : datetime.datetime, optional
        The time the version was first deployed.
    first_deployed_by : UserMetadata, optional
        The user who first deployed the version.
    created_by : UserMetadata, optional
        The user who created the deployment.
    prediction_environment : DeploymentPredictionEnvironment, optional
        The prediction environment of the deployment.
    """

    _converter = t.Dict(
        {
            t.Key("id"): t.String,
            t.Key("currently_deployed"): t.Bool,
            t.Key("registered_model_version"): t.Int,
            t.Key("is_challenger"): t.Bool,
            t.Key("status"): t.String,
            t.Key("label", optional=True): t.Or(t.String, t.Null),
            t.Key("first_deployed_at", optional=True): t.Or(
                t.String() >> dateutil.parser.parse, t.Null
            ),
            t.Key("first_deployed_by", optional=True): t.Or(t.Dict().allow_extra("*"), t.Null),
            t.Key("created_by", optional=True): t.Or(t.Dict().allow_extra("*"), t.Null),
            t.Key("prediction_environment", optional=True): t.Or(t.Dict().allow_extra("*"), t.Null),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: str,
        currently_deployed: bool,
        registered_model_version: int,
        is_challenger: bool,
        status: str,
        label: Optional[str] = None,
        first_deployed_at: Optional[datetime.datetime] = None,
        first_deployed_by: Optional[UserMetadata] = None,
        created_by: Optional[UserMetadata] = None,
        prediction_environment: Optional[DeploymentPredictionEnvironment] = None,
    ):
        self.id = id
        self.currently_deployed = currently_deployed
        self.registered_model_version = registered_model_version
        self.is_challenger = is_challenger
        self.status = status
        self.label = label
        self.first_deployed_at = first_deployed_at
        self.first_deployed_by = first_deployed_by
        self.created_by = created_by
        self.prediction_environment = prediction_environment



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import List, Optional, Union

from typing_extensions import TypedDict

from datarobot.enums import SOURCE_TYPE


class RocCurveEstimatedMetric(TypedDict):
    """Typed dict for estimated metric"""

    accuracy: float
    f1_score: float
    false_negative_score: int
    true_negative_score: int
    true_negative_rate: float
    matthews_correlation_coefficient: float
    true_positive_score: int
    positive_predictive_value: float
    false_positive_score: int
    false_positive_rate: float
    negative_predictive_value: float
    true_positive_rate: float
    threshold: float


class AnomalyAssessmentRecordMetadata(TypedDict):
    """Typed dict for record metadata"""

    record_id: str
    project_id: str
    model_id: str
    backtest: Union[str, int]
    source: SOURCE_TYPE
    series_id: Optional[str]


class AnomalyAssessmentPreviewBin(TypedDict):
    """Typed dict for preview bin"""

    avg_predicted: Optional[float]
    max_predicted: Optional[float]
    start_date: str
    end_date: str
    frequency: int


class ShapleyFeatureContribution(TypedDict):
    """Typed dict for shapley feature contribution"""

    feature_value: str
    strength: float
    feature: str


class AnomalyAssessmentDataPoint(TypedDict):
    """Typed dict for data points"""

    shap_explanation: Optional[List[ShapleyFeatureContribution]]
    timestamp: str
    prediction: float


class RegionExplanationsData(TypedDict):
    """Typed dict for region explanations"""

    explanations: List[AnomalyAssessmentDataPoint]
    shap_base_value: Optional[float]


class Schedule(TypedDict):
    day_of_week: List[Union[int, str]]
    month: List[Union[int, str]]
    hour: List[Union[int, str]]
    minute: List[Union[int, str]]
    day_of_month: List[Union[int, str]]



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import List, Optional, TYPE_CHECKING

import trafaret as t

from datarobot._compat import Int, String
from datarobot.enums import FEATURE_ASSOCIATION_METRIC, FEATURE_ASSOCIATION_TYPE
from datarobot.models.api_object import APIObject
from datarobot.models.status_check_job import StatusCheckJob

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    class Strength(TypedDict):
        feature1: str
        feature2: str
        statistic: float

    class Feature(TypedDict):
        cluster_sort_index: Optional[int]
        cluster_name: Optional[str]
        cluster_id: Optional[int]
        feature: str
        strength_sort_index: int
        alphabetic_sort_index: int
        importance_sort_index: int

    class FeatureAssociationMatrixDict(TypedDict):
        strengths: Optional[List[Strength]]
        features: Optional[List[Feature]]


class FeatureAssociationMatrix(APIObject):
    """
    Feature association statistics for a project.

    .. note::
        Projects created prior to v2.17 are not supported by this feature.

    Attributes
    ----------
    project_id : str
        Id of the associated project.
    strengths : list of dict
        Pairwise statistics for the available features as structured below.
    features : list of dict
        Metadata for each feature and where it goes in the matrix.

    Examples
    --------
    .. code-block:: python

        import datarobot as dr

        # retrieve feature association matrix
        feature_association_matrix = dr.FeatureAssociationMatrix.get(project_id)
        feature_association_matrix.strengths
        feature_association_matrix.features

        # retrieve feature association matrix for a metric, association type or a feature list
        feature_association_matrix = dr.FeatureAssociationMatrix.get(
            project_id,
            metric=enums.FEATURE_ASSOCIATION_METRIC.SPEARMAN,
            association_type=enums.FEATURE_ASSOCIATION_TYPE.CORRELATION,
            featurelist_id=featurelist_id,
        )
    """

    _path = "projects/{}/featureAssociationMatrix/"
    _association_strength = t.Dict(
        {
            t.Key("feature1"): String(allow_blank=True),
            t.Key("feature2"): String(allow_blank=True),
            t.Key("statistic"): t.Float(),
        }
    )
    _association_feature = t.Dict(
        {
            t.Key("cluster_sort_index", optional=True): Int(),
            t.Key("cluster_name", optional=True): String(),
            t.Key("cluster_id", optional=True): Int(),
            t.Key("feature"): String(allow_blank=True),
            t.Key("strength_sort_index"): Int(),
            t.Key("alphabetic_sort_index"): Int(),
            t.Key("importance_sort_index"): Int(),
        }
    )
    _converter = t.Dict(
        {
            t.Key("strengths"): t.List(_association_strength),
            t.Key("features"): t.List(_association_feature),
        }
    )
    _query_param_validator = t.Dict(
        {
            t.Key("metric", optional=True): t.Enum(*FEATURE_ASSOCIATION_METRIC.ALL),
            t.Key("type", optional=True): t.Enum(*FEATURE_ASSOCIATION_TYPE.ALL),
            t.Key("featurelistId", optional=True): t.Or(String(), t.Null()),
        }
    )

    def __init__(
        self,
        strengths: Optional[List[Strength]] = None,
        features: Optional[List[Feature]] = None,
        project_id: Optional[str] = None,
    ) -> None:
        self.strengths = strengths
        self.features = features
        self.project_id = project_id

    def __repr__(self) -> str:
        return "{}(project_id={}, strengths={}, features={})".format(
            self.__class__.__name__,
            self.project_id,
            self.strengths,
            self.features,
        )

    @classmethod
    def get(
        cls,
        project_id: str,
        metric: Optional[str] = None,
        association_type: Optional[str] = None,
        featurelist_id: Optional[str] = None,
    ) -> FeatureAssociationMatrix:
        """
        Get feature association statistics.

        Parameters
        ----------
        project_id : str
            Id of the project that contains the requested associations.
        metric : enums.FEATURE_ASSOCIATION_METRIC
            The name of a metric to get pairwise data for. Since 'v2.19' this is optional and
            defaults to `enums.FEATURE_ASSOCIATION_METRIC.MUTUAL_INFO`.
        association_type : enums.FEATURE_ASSOCIATION_TYPE
            The type of dependence for the data. Since 'v2.19' this is optional and defaults to
            `enums.FEATURE_ASSOCIATION_TYPE.ASSOCIATION`.
        featurelist_id : str or None
            Optional, the feature list to lookup FAM data for. By default, depending on the type of
            the project "Informative Features" or "Timeseries Informative Features" list will be
            used.
            (New in version v2.19)

        Returns
        -------
        FeatureAssociationMatrix
            Feature association pairwise metric strength data, feature clustering data, and
            ordering data for Feature Association Matrix visualization.
        """
        String().check(project_id)
        params = {}
        if metric:
            params["metric"] = metric
        if association_type:
            params["type"] = association_type
        if featurelist_id:
            params["featurelistId"] = featurelist_id
        cls._query_param_validator.check(params)

        url = cls._path.format(project_id)
        response = cls._client.get(url, params=params)
        feature_association_matrix = cls.from_server_data(response.json())
        # FAM public API doesn't include project_id so lets populate it
        feature_association_matrix.project_id = project_id
        return feature_association_matrix

    @classmethod
    def create(cls, project_id: str, featurelist_id: str) -> StatusCheckJob:
        """Compute the Feature Association Matrix for a Feature List

        Parameters
        ----------
        project_id : str
            The ID of the project that the feature list belongs to.
        featurelist_id : str
            The ID of the feature list for which insights are requested.

        Returns
        -------
        status_check_job : StatusCheckJob
            Object contains all needed logic for a periodical status check of an async job.
        """

        url = cls._path.format(project_id)
        payload = {"featurelistId": featurelist_id}
        resp = cls._client.post(url, data=payload)
        return StatusCheckJob.from_response(resp, response_type=FeatureAssociationMatrix)

    def to_dict(self) -> FeatureAssociationMatrixDict:
        return {"strengths": self.strengths, "features": self.features}



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from io import IOBase
from typing import Optional, Union

import pandas as pd
import trafaret as t

from datarobot._compat import String
from datarobot.enums import FileLocationType, LocalSourceType
from datarobot.errors import InvalidUsageError
from datarobot.models.api_object import APIObject
from datarobot.models.dataset import Dataset
from datarobot.models.project import PredictionDataset, Project
from datarobot.utils.source import parse_source_type

from ..enums import DEFAULT_MAX_WAIT
from ..utils.waiters import wait_for_async_resolution


class DataEngineQueryGenerator(APIObject):
    """DataEngineQueryGenerator is used to set up time series data prep.

    .. versionadded:: v2.27

    Attributes
    ----------
    id: str
        id of the query generator
    query: str
        text of the generated Spark SQL query
    datasets: list(QueryGeneratorDataset)
        datasets associated with the query generator
    generator_settings: QueryGeneratorSettings
        the settings used to define the query
    generator_type: str
        "TimeSeries" is the only supported type
    """

    _path = "dataEngineQueryGenerators/"
    _converter = t.Dict(
        {
            t.Key("datasets", optional=True): t.List(
                t.Dict(
                    {
                        t.Key("alias"): String,
                        t.Key("dataset_id"): String,
                        t.Key("dataset_version_id", optional=True): String,
                    }
                )
            ),
            t.Key("generator_settings"): t.Dict(
                {
                    t.Key("datetime_partition_column"): String,
                    t.Key("time_unit"): String,
                    t.Key("time_step"): t.Int,
                    t.Key("default_numeric_aggregation_method"): String,
                    t.Key("default_categorical_aggregation_method"): String,
                    t.Key("target", optional=True): String,
                    t.Key("multiseries_id_columns", optional=True): t.List(String),
                    t.Key("default_text_aggregation_method", optional=True): String,
                    t.Key("start_from_series_min_datetime", optional=True): t.Bool,
                    t.Key("end_to_series_max_datetime", optional=True): t.Bool,
                }
            ).allow_extra("*"),
            t.Key("generator_type"): String,
            t.Key("id"): String,
            t.Key("query"): String,
        }
    ).allow_extra("*")

    def __init__(self, **generator_kwargs):
        self.datasets = generator_kwargs["datasets"]
        self.generator_type = generator_kwargs["generator_type"]
        self.generator_settings = generator_kwargs["generator_settings"]
        self.id = generator_kwargs["id"]
        self.query = generator_kwargs["query"]

    def __repr__(self):
        return (
            "{}(generator_id={}, datasets={}, generator_type={}, " "generator_settings={}, query={}"
        ).format(
            self.__class__.__name__,
            self.id,
            self.datasets,
            self.generator_type,
            self.generator_settings,
            self.query,
        )

    @classmethod
    def create(cls, generator_type, datasets, generator_settings):
        """Creates a query generator entity.

        .. versionadded:: v2.27

        Parameters
        ----------
        generator_type : str
            Type of data engine query generator
        datasets : List[QueryGeneratorDataset]
            Source datasets in the Data Engine workspace.
        generator_settings : dict
            Data engine generator settings of the given `generator_type`.

        Returns
        -------
        query_generator : DataEngineQueryGenerator
            The created generator

        Examples
        --------
        .. code-block:: python

            import datarobot as dr
            from datarobot.models.data_engine_query_generator import (
               QueryGeneratorDataset,
               QueryGeneratorSettings,
            )
            dataset = QueryGeneratorDataset(
               alias='My_Awesome_Dataset_csv',
               dataset_id='61093144cabd630828bca321',
               dataset_version_id=1,
            )
            settings = QueryGeneratorSettings(
               datetime_partition_column='date',
               time_unit='DAY',
               time_step=1,
               default_numeric_aggregation_method='sum',
               default_categorical_aggregation_method='mostFrequent',
            )
            g = dr.DataEngineQueryGenerator.create(
               generator_type='TimeSeries',
               datasets=[dataset],
               generator_settings=settings,
            )
            g.id
            >>>'54e639a18bd88f08078ca831'
            g.generator_type
            >>>'TimeSeries'
        """
        data = {
            "generator_type": generator_type,
            "datasets": [d.to_dict() for d in datasets],
            "generator_settings": generator_settings.to_dict(),
        }
        response = cls._client.post(cls._path, data=data)
        finished_url = wait_for_async_resolution(cls._client, response.headers["Location"])
        finished_response = cls._client.get(finished_url)

        return cls.from_server_data(finished_response.json())

    @classmethod
    def get(cls, generator_id):
        """Gets information about a query generator.

        Parameters
        ----------
        generator_id : str
            The identifier of the query generator you want to load.

        Returns
        -------
        query_generator : DataEngineQueryGenerator
            The queried generator

        Examples
        --------
        .. code-block:: python

            import datarobot as dr
            g = dr.DataEngineQueryGenerator.get(generator_id='54e639a18bd88f08078ca831')
            g.id
            >>>'54e639a18bd88f08078ca831'
            g.generator_type
            >>>'TimeSeries'
        """
        path = f"{cls._path}{generator_id}/"
        return cls.from_location(path)

    def create_dataset(self, dataset_id=None, dataset_version_id=None, max_wait=DEFAULT_MAX_WAIT):
        """
        A blocking call that creates a new Dataset from the query generator.
        Returns when the dataset has been successfully processed. If optional
        parameters are not specified the query is applied to the dataset_id
        and dataset_version_id stored in the query generator. If specified they
        will override the stored dataset_id/dataset_version_id, i.e. to prep a
        prediction dataset.

        Parameters
        ----------
        dataset_id: str, optional
            The id of the unprepped dataset to apply the query to
        dataset_version_id: str, optional
            The version_id of the unprepped dataset to apply the query to

        Returns
        -------
        response: Dataset
            The Dataset created from the query generator
        """
        return Dataset.create_from_query_generator(
            self.id, dataset_id, dataset_version_id, max_wait
        )

    def prepare_prediction_dataset_from_catalog(
        self,
        project_id: str,
        dataset_id: str,
        dataset_version_id: Optional[str] = None,
        max_wait: Optional[int] = DEFAULT_MAX_WAIT,
        relax_known_in_advance_features_check: Optional[bool] = None,
    ) -> PredictionDataset:
        """Apply time series data prep to a catalog dataset and upload it to the project
        as a PredictionDataset.

        .. versionadded:: v3.1

        Parameters
        ----------
        project_id : str
            The id of the project to which you upload the prediction dataset.
        dataset_id : str
            The identifier of the dataset.
        dataset_version_id : str, optional
            The version id of the dataset to use.
        max_wait : int, optional
            Optional, the maximum number of seconds to wait before giving up.
        relax_known_in_advance_features_check : bool, optional
            For time series projects only. If True, missing values in the
            known in advance features are allowed in the forecast window at the prediction time.
            If omitted or False, missing values are not allowed.

        Returns
        -------
        dataset : PredictionDataset
            The newly uploaded dataset.
        """
        prediction_dataset = self.create_dataset(dataset_id, dataset_version_id, max_wait)
        project = Project.get(project_id)

        return project.upload_dataset_from_catalog(
            dataset_id=prediction_dataset.id,
            dataset_version_id=prediction_dataset.version_id,
            max_wait=max_wait,
            relax_known_in_advance_features_check=relax_known_in_advance_features_check,
        )

    def prepare_prediction_dataset(
        self,
        sourcedata: Union[str, pd.DataFrame, IOBase],
        project_id: str,
        max_wait: Optional[int] = DEFAULT_MAX_WAIT,
        relax_known_in_advance_features_check: Optional[bool] = None,
    ) -> PredictionDataset:
        """Apply time series data prep and upload the PredictionDataset to the project.

        .. versionadded:: v3.1

        Parameters
        ----------
        sourcedata : str, file or pandas.DataFrame
            Data to be used for predictions. If it is a string, it can be either a path to a local file,
            or raw file content. If using a file on disk, the filename must consist of ASCII
            characters only.
        project_id : str
            The id of the project to which you upload the prediction dataset.
        max_wait : int, optional
            The maximum number of seconds to wait for the uploaded dataset to be processed before
            raising an error.
        relax_known_in_advance_features_check : bool, optional
            For time series projects only. If True, missing values in the
            known in advance features are allowed in the forecast window at the prediction time.
            If omitted or False, missing values are not allowed.
        Returns
        -------
        dataset : PredictionDataset
            The newly uploaded dataset.

        Raises
        ------
        InputNotUnderstoodError
            Raised if ``sourcedata`` isn't one of supported types.
        AsyncFailureError
            Raised if polling for the status of an async process resulted in a response with an
            unsupported status code.
        AsyncProcessUnsuccessfulError
            Raised if project creation was unsuccessful (i.e. the server reported an error in
            uploading the dataset).
        AsyncTimeoutError
            Raised if processing the uploaded dataset took more time than specified
            by the ``max_wait`` parameter.
        """
        source_type = parse_source_type(sourcedata)
        if source_type not in (
            FileLocationType.PATH,
            LocalSourceType.DATA_FRAME,
            LocalSourceType.FILELIKE,
        ):
            raise InvalidUsageError(
                f"Unable to parse source ({sourcedata}) as filepath, DataFrame, or file."
            )

        dataset = Dataset.upload(sourcedata)
        prediction_dataset = self.create_dataset(dataset.id, dataset.version_id, max_wait)
        project = Project.get(project_id)
        return project.upload_dataset_from_catalog(
            dataset_id=prediction_dataset.id,
            dataset_version_id=prediction_dataset.version_id,
            max_wait=max_wait,
            relax_known_in_advance_features_check=relax_known_in_advance_features_check,
        )


class QueryGeneratorSettings:
    """
    A QueryGenerator settings to be used to create DataEngineQueryGenerator entity.
    .. versionadded:: v2.27
    Attributes
    ----------
    datetime_partition_column : str
        The date column that will be used as a datetime partition column
        in time series project.
    time_unit : str
        Indicates which unit is a basis for time steps of the output dataset.
    time_step : int
        Number of time steps for the output dataset.
    default_numeric_aggregation_method : str
        Default aggregation method used for numeric feature.
    default_categorical_aggregation_method : str
        Default aggregation method used for categorical feature.
    target : str, optional
        The name of target for the output dataset.
    multiseries_id_columns : list[str], optional
        An array with the names of columns identifying the series to which
        row of the output dataset belongs. Currently, only one multiseries
        ID column is supported.
    default_text_aggregation_method : str, optional
        Default aggregation method used for text feature.
    start_from_series_min_datetime : bool, optional
        A boolean value indicating whether post-aggregated series starts from series minimum
        datetime or global minimum datetime.
    end_to_series_max_datetime : bool, optional
        A boolean value indicating whether generates post-aggregated series up to series
        maximum datetime or global maximum datetime.
    Examples
    --------
    .. code-block:: python
        from datarobot.models.data_engine_query_generator import QueryGeneratorSettings
        query_generator_settings = QueryGeneratorSettings(
            datetime_partition_column='date',
            time_unit='DAY',
            time_step=1,
            default_numeric_aggregation_method='sum',
            default_categorical_aggregation_method='mostFrequent',
        )
    """

    def __init__(
        self,
        datetime_partition_column,
        time_unit,
        time_step,
        default_numeric_aggregation_method,
        default_categorical_aggregation_method,
        target=None,
        multiseries_id_columns=None,
        default_text_aggregation_method=None,
        start_from_series_min_datetime=True,
        end_to_series_max_datetime=True,
    ):
        self.datetime_partition_column = datetime_partition_column
        self.time_unit = time_unit
        self.time_step = time_step
        self.default_numeric_aggregation_method = default_numeric_aggregation_method
        self.default_categorical_aggregation_method = default_categorical_aggregation_method
        self.target = target
        self.multiseries_id_columns = multiseries_id_columns
        self.default_text_aggregation_method = default_text_aggregation_method
        self.start_from_series_min_datetime = start_from_series_min_datetime
        self.end_to_series_max_datetime = end_to_series_max_datetime

    def to_dict(self):
        return {
            "datetime_partition_column": self.datetime_partition_column,
            "time_unit": self.time_unit,
            "time_step": self.time_step,
            "default_numeric_aggregation_method": self.default_numeric_aggregation_method,
            "default_categorical_aggregation_method": self.default_categorical_aggregation_method,
            "target": self.target,
            "multiseries_id_columns": self.multiseries_id_columns,
            "default_text_aggregation_method": self.default_text_aggregation_method,
            "start_from_series_min_datetime": self.start_from_series_min_datetime,
            "end_to_series_max_datetime": self.end_to_series_max_datetime,
        }

    def to_payload(self):
        return {
            "datetimePartitionColumn": self.datetime_partition_column,
            "timeUnit": self.time_unit,
            "timeStep": self.time_step,
            "defaultNumericAggregationMethod": self.default_numeric_aggregation_method,
            "defaultCategoricalAggregationMethod": self.default_categorical_aggregation_method,
            "target": self.target,
            "multiseriesIdColumns": self.multiseries_id_columns,
            "defaultTextAggregationMethod": self.default_text_aggregation_method,
            "startFromSeriesMinDatetime": self.start_from_series_min_datetime,
            "endToSeriesMaxDatetime": self.end_to_series_max_datetime,
        }


class QueryGeneratorDataset:
    """
    A QueryGenerator dataset to be used to create DataEngineQueryGenerator entity.
    .. versionadded:: v2.27
    Attributes
    ----------
    alias : str
        The name for the dataset.
    dataset_id : str
        The identifier of the dataset.
    dataset_version_id : str, optional
        The version id of the dataset to use.
    Examples
    --------
    .. code-block:: python
        from datarobot.models.data_engine_query_generator import QueryGeneratorDataset
        query_generator_dataset = QueryGeneratorDataset(
            alias='My_Awesome_Dataset_csv',
            dataset_id='61093144cabd630828bca321',
            dataset_version_id=1,
        )
    """

    def __init__(self, alias, dataset_id, dataset_version_id=None):
        self.alias = alias
        self.dataset_id = dataset_id
        self.dataset_version_id = dataset_version_id

    def to_dict(self):
        return {
            "alias": self.alias,
            "dataset_id": self.dataset_id,
            "dataset_version_id": self.dataset_version_id,
        }

    def to_payload(self):
        return {
            "alias": self.alias,
            "datasetId": self.dataset_id,
            "datasetVersionId": self.dataset_version_id,
        }



# flake8: noqa
from .external_confusion_chart import ExternalConfusionChart
from .external_dataset_residuals import ExternalResidualsChart
from .external_lift_chart import ExternalLiftChart
from .external_multiclass_lift_chart import ExternalMulticlassLiftChart
from .external_roc_curve import ExternalRocCurve
from .external_scores import ExternalScores



#
# Copyright 2021-2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

import contextlib
import json
import os
from typing import Any, Dict, Generator, List, Optional, Set, Tuple, Union

from requests_toolbelt import MultipartEncoder
import trafaret as t
from typing_extensions import TypedDict

from datarobot._compat import String
from datarobot.models.api_object import APIObject
from datarobot.models.runtime_parameters import RuntimeParameter, RuntimeParameterValue
from datarobot.models.types import Schedule
from datarobot.utils import camelize
from datarobot.utils.pagination import unpaginate


class JobFileItem(APIObject):
    """A file item attached to a DataRobot job.

    .. versionadded:: v3.4

    Attributes
    ----------
    id: str
        The ID of the file item.
    file_name: str
        The name of the file item.
    file_path: str
        The path of the file item.
    file_source: str
        The source of the file item.
    created_at: str
        ISO-8601 formatted timestamp of when the version was created.
    """

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("file_name"): String(),
            t.Key("file_path"): String(),
            t.Key("file_source"): String(),
            t.Key("created") >> "created_at": String(),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(
        self,
        id: str,
        file_name: str,
        file_path: str,
        file_source: str,
        created_at: str,
    ) -> None:
        self.id = id
        self.file_name = file_name
        self.file_path = file_path
        self.file_source = file_source
        self.created_at = created_at


class JobFileItemType(TypedDict):
    id: str
    file_name: str
    file_path: str
    file_source: str
    created_at: str


class Job(APIObject):
    """A DataRobot job.

    .. versionadded:: v3.4

    Attributes
    ----------
    id: str
        The ID of the job.
    name: str
        The name of the job.
    created_at: str
        ISO-8601 formatted timestamp of when the version was created
    items: List[JobFileItem]
        A list of file items attached to the job.
    description: str, optional
        A job description.
    environment_id: str, optional
        The ID of the environment to use with the job.
    environment_version_id: str, optional
        The ID of the environment version to use with the job.
    """

    _path = "customJobs/"

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("name"): String(),
            t.Key("created") >> "created_at": String(),
            t.Key("items"): t.List(JobFileItem.schema),
            t.Key("description", optional=True): t.Or(
                String(max_length=10000, allow_blank=True), t.Null()
            ),
            t.Key("environment_id", optional=True): t.Or(String(), t.Null()),
            t.Key("environment_version_id", optional=True): t.Or(String(), t.Null()),
            t.Key("entry_point", optional=True): t.Or(String(), t.Null()),
            t.Key("runtime_parameters", optional=True): t.List(RuntimeParameter.schema),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(
        self,
        id: str,
        name: str,
        created_at: str,
        items: List[JobFileItemType],
        description: Optional[str] = None,
        environment_id: Optional[str] = None,
        environment_version_id: Optional[str] = None,
        entry_point: Optional[str] = None,
        runtime_parameters: Optional[List[RuntimeParameter]] = None,
    ) -> None:
        self.id = id
        self.description = description
        self.name = name
        self.created_at = created_at
        self.items = [JobFileItem(**data) for data in items]

        self.environment_id = environment_id
        self.environment_version_id = environment_version_id
        self.entry_point = entry_point

        self.runtime_parameters = (
            [RuntimeParameter(**param) for param in runtime_parameters]  # type: ignore[arg-type]
            if runtime_parameters
            else None
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name or self.id!r})"

    def _update_values(self, new_response: Job) -> None:
        fields: Set[str] = self._fields()  # type: ignore[no-untyped-call]
        for attr in fields:
            new_value = getattr(new_response, attr)
            setattr(self, attr, new_value)

    @classmethod
    def _job_path(cls, job_id: str) -> str:
        return f"{cls._path}{job_id}/"

    @classmethod
    def create(
        cls,
        name: str,
        environment_id: Optional[str] = None,
        environment_version_id: Optional[str] = None,
        folder_path: Optional[str] = None,
        files: Optional[Union[List[Tuple[str, str]], List[str]]] = None,
        file_data: Optional[Dict[str, str]] = None,
        runtime_parameter_values: Optional[List[RuntimeParameterValue]] = None,
    ) -> Job:
        """Create a job.

        .. versionadded:: v3.4

        Parameters
        ----------
        name: str
            The name of the job.
        environment_id: Optional[str]
            The environment ID to use for job runs.
            The ID must be specified in order to run the job.
        environment_version_id: Optional[str]
            The environment version ID to use for job runs.
            If not specified, the latest version of the execution environment will be used.
        folder_path: Optional[str]
            The path to a folder containing files to be uploaded.
            Each file in the folder is uploaded under path relative
            to a folder path.
        files: Optional[Union[List[Tuple[str, str]], List[str]]]
            The files to be uploaded to the job.
            The files can be defined in 2 ways:
            1. List of tuples where 1st element is the local path of the file to be uploaded
            and the 2nd element is the file path in the job file system.
            2. List of local paths of the files to be uploaded.
            In this case files are added to the root of the model file system.
        file_data: Optional[Dict[str, str]]
            The files content to be uploaded to the job.
            Defined as a dictionary where keys are the file paths in the job file system.
            and values are the files content.
        runtime_parameter_values: Optional[List[RuntimeParameterValue]]
            Additional parameters to be injected into a model at runtime. The fieldName
            must match a fieldName that is listed in the runtimeParameterDefinitions section
            of the model-metadata.yaml file.

        Returns
        -------
        Job
            created job

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        upload_data: List[Tuple[str, Any]] = [
            ("name", name),
        ]

        if environment_id:
            upload_data.append(
                ("environmentId", environment_id),
            )
        if environment_version_id:
            upload_data.append(
                ("environmentVersionId", environment_version_id),
            )

        if runtime_parameter_values is not None:
            upload_data.append(
                (
                    "runtimeParameterValues",
                    json.dumps(
                        [
                            {camelize(k): v for k, v in param.to_dict().items()}
                            for param in runtime_parameter_values
                        ]
                    ),
                )
            )

        with cls._process_files_upload(folder_path, files, file_data) as file_upload_data:
            encoder = MultipartEncoder(fields=upload_data + file_upload_data)
            headers = {"Content-Type": encoder.content_type}
            response = cls._client.request("post", cls._path, data=encoder, headers=headers)

        return cls.from_server_data(response.json())

    @staticmethod
    def _verify_folder_path(folder_path: Optional[str]) -> None:
        if folder_path and not os.path.exists(folder_path):
            raise ValueError(f"The folder: {folder_path} does not exist.")

    @classmethod
    @contextlib.contextmanager
    def _process_files_upload(
        cls,
        folder_path: Optional[str] = None,
        files: Optional[Union[List[Tuple[str, str]], List[str]]] = None,
        file_data: Optional[Dict[str, str]] = None,
    ) -> Generator[List[Tuple[str, Any]], None, None]:
        """Process file-related argument and prepare them for the uploading."""
        upload_data: List[Tuple[str, Any]] = []

        cls._verify_folder_path(folder_path)

        with contextlib.ExitStack() as stack:
            if folder_path:
                for dir_name, _, file_names in os.walk(folder_path):
                    for file_name in file_names:
                        file_path = os.path.join(dir_name, file_name)
                        file = stack.enter_context(open(file_path, "rb"))

                        upload_data.append(("file", (os.path.basename(file_path), file)))
                        upload_data.append(("filePath", os.path.relpath(file_path, folder_path)))

            if files:
                # `paths` is a dict of (target file path, source file path) pairs, where
                # - target file path - is the file path on the job filesystem
                # - source file path - is the file path of the file to be uploaded
                paths: List[Tuple[str, str]]
                if isinstance(files[0], tuple):
                    paths = files  # type: ignore[assignment]
                else:
                    paths = []
                    for p in files:
                        source_file_path: str = p  # type: ignore[assignment]
                        paths.append((source_file_path, os.path.basename(source_file_path)))

                for source_file_path, target_file_path in paths:
                    source_file = stack.enter_context(open(source_file_path, "rb"))

                    upload_data.append(("file", (os.path.basename(target_file_path), source_file)))
                    upload_data.append(("filePath", target_file_path))

            if file_data:
                for target_file_path, content in file_data.items():
                    upload_data.append(("file", (os.path.basename(target_file_path), content)))
                    upload_data.append(("filePath", target_file_path))

            yield upload_data

    @classmethod
    def list(cls) -> List[Job]:
        """List jobs.

        .. versionadded:: v3.4

        Returns
        -------
        List[Job]
            a list of jobs

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        data = unpaginate(cls._path, None, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def get(cls, job_id: str) -> Job:
        """Get job by id.

        .. versionadded:: v3.4

        Parameters
        ----------
        job_id: str
            The ID of the job.

        Returns
        -------
        Job
            retrieved job

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        path = cls._job_path(job_id)
        return cls.from_location(path)

    def update(
        self,
        name: Optional[str] = None,
        entry_point: Optional[str] = None,
        environment_id: Optional[str] = None,
        environment_version_id: Optional[str] = None,
        description: Optional[str] = None,
        folder_path: Optional[str] = None,
        files: Optional[Union[List[Tuple[str, str]], List[str]]] = None,
        file_data: Optional[Dict[str, str]] = None,
        runtime_parameter_values: Optional[List[RuntimeParameterValue]] = None,
    ) -> None:
        """Update job properties.

        .. versionadded:: v3.4

        Parameters
        ----------
        name: str
            The job name.
        entry_point: Optional[str]
            The job file item ID to use as an entry point of the job.
        environment_id: Optional[str]
            The environment ID to use for job runs.
            Must be specified in order to run the job.
        environment_version_id: Optional[str]
            The environment version ID to use for job runs.
            If not specified, the latest version of the execution environment will be used.
        description: str
            The job description.
        folder_path: Optional[str]
            The path to a folder containing files to be uploaded.
            Each file in the folder is uploaded under path relative
            to a folder path.
        files: Optional[Union[List[Tuple[str, str]], List[str]]]
            The files to be uploaded to the job.
            The files can be defined in 2 ways:
            1. List of tuples where 1st element is the local path of the file to be uploaded
            and the 2nd element is the file path in the job file system.
            2. List of local paths of the files to be uploaded.
            In this case files are added to the root of the job file system.
        file_data: Optional[Dict[str, str]]
            The files content to be uploaded to the job.
            Defined as a dictionary where keys are the file paths in the job file system.
            and values are the files content.
        runtime_parameter_values: Optional[List[RuntimeParameterValue]]
            Additional parameters to be injected into a model at runtime. The fieldName
            must match a fieldName that is listed in the runtimeParameterDefinitions section
            of the model-metadata.yaml file.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        path = self._job_path(self.id)

        upload_data: List[Tuple[str, Any]] = []

        if name:
            upload_data.append(("name", name))
        if entry_point:
            upload_data.append(("entryPoint", entry_point))
        if environment_id:
            upload_data.append(("environmentId", environment_id))
        if environment_version_id:
            upload_data.append(("environmentVersionId", environment_version_id))
        if description:
            upload_data.append(("description", description))
        if runtime_parameter_values:
            upload_data.append(
                (
                    "runtimeParameterValues",
                    json.dumps(
                        [
                            {camelize(k): v for k, v in param.to_dict().items()}
                            for param in runtime_parameter_values
                        ]
                    ),
                )
            )

        with self._process_files_upload(folder_path, files, file_data) as file_upload_data:
            encoder = MultipartEncoder(fields=upload_data + file_upload_data)
            headers = {"Content-Type": encoder.content_type}
            response = self._client.request("patch", path, data=encoder, headers=headers)

        data = response.json()
        new_version = Job.from_server_data(data)
        self._update_values(new_version)

    def delete(self) -> None:
        """Delete job.

        .. versionadded:: v3.4

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        url = f"{self._path}{self.id}/"
        self._client.delete(url)

    def refresh(self) -> None:
        """Update job with the latest data from server.

        .. versionadded:: v3.4

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """

        new_object = self.get(self.id)
        self._update_values(new_object)

    @classmethod
    def create_from_custom_metric_gallery_template(
        cls,
        template_id: str,
        name: str,
        description: Optional[str] = None,
        sidecar_deployment_id: Optional[str] = None,
    ) -> Job:
        """Create a job from a custom metric gallery template.

        Parameters
        ----------
        template_id: str
            ID of the template.
        name: str
            Name of the job.
        description: Optional[str]
            Description of the job.
        sidecar_deployment_id: Optional[str]
            ID of the sidecar deployment. Only relevant for templates that use sidecar deployments.

        Returns
        -------
        Job
            retrieved job

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        payload = {"name": name, "templateId": template_id}
        if description:
            payload["description"] = description
        if sidecar_deployment_id:
            payload["sidecarDeploymentId"] = sidecar_deployment_id
        response = cls._client.post(
            f"{cls._path}fromHostedCustomMetricGalleryTemplate/", data=payload
        )
        return cls.from_server_data(response.json())

    def list_schedules(self) -> List[JobSchedule]:
        """List schedules for the job.

        Returns
        -------
        List[JobSchedule]
            a list of schedules for the job.
        """
        url = f"customJobs/{self.id}/schedules/"
        data = unpaginate(url, None, self._client)
        return [JobSchedule.from_server_data(item) for item in data]


class JobUser(TypedDict, total=False):
    id: str
    username: str
    first_name: Optional[str]
    last_name: Optional[str]


class JobDeployment(TypedDict, total=False):
    id: str
    name: str
    description: Optional[str]


class JobSchedule(APIObject):
    """A job schedule.

    .. versionadded:: v3.5

    Attributes
    ----------
    id: str
        The ID of the job schedule.
    custom_job_id: str
        The ID of the custom job.
    updated_at: str
        ISO-8601 formatted timestamp of when the schedule was updated.
    updated_by: Dict[str, Any]
        The user who updated the schedule.
    created_at: str
        ISO-8601 formatted timestamp of when the schedule was created.
    created_by: Dict[str, Any]
        The user who created the schedule.
    scheduled_job_id: str
        The ID of the scheduled job.
    deployment: Dict[str, Any]
        The deployment of the scheduled job.
    schedule: Schedule
        The schedule of the job.
    parameter_overrides: List[RuntimeParameterValue]
        The parameter overrides for this schedule.

    """

    _user = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("username"): String(),
            t.Key("first_name"): t.Or(String(), t.Null()),
            t.Key("last_name"): t.Or(String(), t.Null()),
        }
    ).allow_extra("*")
    _deployment = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("name"): String(),
            t.Key("description"): t.Or(String(), t.Null()),
        }
    ).allow_extra("*")

    _parameter_overrides = t.Dict(
        {
            t.Key("field_name"): t.String(),
            t.Key("value", optional=True): t.Or(t.String(), t.Int(), t.Float(), t.Bool(), t.Null()),
            t.Key("type"): t.String(),
        }
    ).allow_extra("*")

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("custom_job_id"): String(),
            t.Key("updated_at"): String(),
            t.Key("updated_by"): _user,
            t.Key("created_at"): String(),
            t.Key("created_by"): _user,
            t.Key("scheduled_job_id"): String(),
            t.Key("deployment", optional=True): _deployment,
            t.Key("schedule", optional=True): t.Dict(
                {
                    t.Key("hour"): t.List(t.Or(t.String(), t.Int())),
                    t.Key("minute"): t.List(t.Or(t.String(), t.Int())),
                    t.Key("day_of_week"): t.List(t.Or(t.String(), t.Int())),
                    t.Key("day_of_month"): t.List(t.Or(t.String(), t.Int())),
                    t.Key("month"): t.List(t.Or(t.String(), t.Int())),
                }
            ).allow_extra("*"),
            t.Key("parameter_overrides", optional=True): t.List(RuntimeParameterValue.schema),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: str,
        custom_job_id: str,
        updated_at: str,
        updated_by: JobUser,
        created_at: str,
        created_by: JobUser,
        scheduled_job_id: str,
        schedule: Optional[Schedule] = None,
        deployment: Optional[JobDeployment] = None,
        parameter_overrides: Optional[List[RuntimeParameterValue]] = None,
    ) -> None:
        """
        Parameters
        ----------
        id: str
            The ID of the job schedule.
        custom_job_id: str
            The ID of the custom job.
        updated_at: str
            ISO-8601 formatted timestamp of when the schedule was updated.
        updated_by: JobUser
            The user who updated the schedule.
        created_at: str
            ISO-8601 formatted timestamp of when the schedule was created.
        created_by: JobUser
            The user who created the schedule.
        scheduled_job_id: str
            The ID of the scheduled job.
        deployment: Optional[JobDeployment]
            The deployment of the scheduled job.
        schedule: Optional[Schedule]
            The schedule of the job.
        parameter_overrides: Optional[List[RuntimeParameterValue]]
            The parameter overrides for this schedule.

        """
        self.id = id
        self.custom_job_id = custom_job_id
        self.updated_at = updated_at
        self.updated_by = updated_by
        self.created_at = created_at
        self.created_by = created_by
        self.scheduled_job_id = scheduled_job_id
        self.deployment = deployment
        self.schedule = schedule
        self.parameter_overrides = parameter_overrides

    def update(
        self,
        schedule: Optional[Schedule] = None,
        parameter_overrides: Optional[List[RuntimeParameterValue]] = None,
    ) -> JobSchedule:
        """Update the job schedule.

        Parameters
        ----------
        schedule: Optional[Schedule]
            The schedule of the job.
        parameter_overrides: Optional[List[RuntimeParameterValue]]
            The parameter overrides for this schedule.
        """
        url = f"customJobs/{self.custom_job_id}/schedules/{self.id}/"
        payload: Dict[str, Any] = {}
        if schedule:
            payload["schedule"] = schedule
        if parameter_overrides:
            payload["parameter_overrides"] = [
                {"fieldName": param.field_name, "value": param.value, "type": param.type}
                for param in parameter_overrides
            ]
        response = self._client.patch(url, data=payload)
        return self.from_server_data(response.json())

    def delete(self) -> None:
        """Delete the job schedule.
        Returns
        -------
        None
        """
        url = f"customJobs/{self.custom_job_id}/schedules/{self.id}/"
        self._client.delete(url)

    @classmethod
    def create(
        cls,
        custom_job_id: str,
        schedule: Schedule,
        parameter_overrides: Optional[List[RuntimeParameterValue]] = None,
    ) -> JobSchedule:
        """Create a job schedule.

        Parameters
        ----------
        custom_job_id: str
            The ID of the custom job.
        schedule: Schedule
            The schedule of the job.
        parameter_overrides: Optional[List[RuntimeParameterValue]]
            The parameter overrides for this schedule.
        """
        url = f"customJobs/{custom_job_id}/schedules/"
        payload: Dict[str, Any] = {"schedule": schedule}
        if parameter_overrides:
            payload["parameter_overrides"] = [
                {"fieldName": param.field_name, "value": param.value, "type": param.type}
                for param in parameter_overrides
            ]
        response = cls._client.post(url, data=payload)
        return cls.from_server_data(response.json())



# flake8: noqa
from .chat import Chat
from .chat_prompt import ChatPrompt
from .comparison_chat import ComparisonChat
from .comparison_prompt import ComparisonPrompt
from .custom_model_llm_validation import CustomModelLLMValidation
from .llm import LLMDefinition
from .llm_blueprint import LLMBlueprint
from .playground import Playground
from .user_limits import UserLimits
from .vector_database import CustomModelVectorDatabaseValidation, VectorDatabase
